/*
  Copyright 2006 by Sean Luke and George Mason University
  Licensed under the Academic Free License version 3.0
  See the file "LICENSE" for more information
*/

package batch;

import sim.engine.*;
import sim.util.*;
import sim.field.continuous.*;
import java.io.*;

public class Engine extends SimState {
    //Initialises files to take data for model-level and individual agent variable records
    static PrintWriter outFile=null;
    static PrintWriter macroOutFile=null;
    
    //Variables for changing parameters over successive runs
    int numRuns;
    double paramMin;
    double paramInc;
    int batchNum = 0;
    
    //Parameters which will be varied in testing stage
    int startNeutros;
    double opsoninLevel;
    double PAMax;
    double PBMax;
    double velocity;
    
    double spike;
    
    //Arrays to hold integer and double data from each Macro agent at subsequent timesteps
    public int macroIntData[][] = new int[2][20];
    public double macroDoubleData[][] = new double[6][20];
    
    //Dimensions of toroidal environment
    public static final double XMIN = 0;
    public static final double XMAX = 200;
    public static final double YMIN = 0;
    public static final double YMAX = 200;
    
    //Temporary variable for macrophage angle of movement
    public double tempAngle;
    
    //Accounting variables for number of each type of Neutro agent in simulation
    public int numNeutros;
    public int numLive;
    public int numApop;
    public int numBound;
    public int numNecrotic;
    
    //(Fixed) number of Macro agents in simulation
    public int numMacros = 20;
    
    //Variables for concentration of signalling molecules and probability that new Neutro agents are introduced
    double CBeta;
    double CAlpha;
    double neutroProb;
    
    //Methods for reading in/out values of variables and parameters in model via GUI interface
    public double getOpsoninLevel() {return opsoninLevel;}
    public void setOpsoninLevel(double value) {
        if(value <= 100 && value >= 0) {
            opsoninLevel = value;
        }
        return;
    }
    public double getCBeta() {return CBeta;}
    public double getCAlpha() {return CAlpha;}
    public double getNeutroProb() {return neutroProb*100;}
    public void setNeutroProb(double value) {
        neutroProb = value;
    }
    public int getNumNeutros() {return numNeutros;}
    public int getLive() {return numLive;}
    public int getApoptotic() {return numApop+numBound;}
    public int getNecrotic() {return numNecrotic;}
    
    //Declare environment type: continuous 2D
    public Continuous2D agentsEnvironment = null;
    
    /** Creates a new instance of Movement */
    public Engine(long seed) {
        super(seed);
    }
    
    /**Sets location of agents according to their coordinates in toroidal environment*/
    public void setObjectLocation( final Agent agent, Double2D location )
        {
        // Toroidal coordinate translation
        double x = (((location.x - XMIN) + (XMAX-XMIN)) % (XMAX-XMIN)) + XMIN;
        double y = (((location.y - YMIN) + (YMAX-YMIN)) % (YMAX-YMIN)) + YMIN;
        
        location = new Double2D( x, y );

        agentsEnvironment.setObjectLocation( agent, location );
        agent.x = location.x;
        agent.y = location.y;
        }
    
    /**Sets parameters to correct start values and allows one parameter to be incremented in consecutive batches of runs*/
    void resetValues() {
        numRuns = 20;
        paramMin = 60;
        paramInc = 10;
        
        startNeutros = 100;
        opsoninLevel = 0.2;
        PAMax = 0.333;
        PBMax = 0.135;
        velocity = 1.0;
        
        numNeutros = 0;
        numApop = 0;
        numLive = 0;
        numNecrotic = 0;
        numBound = 0;
        
        CAlpha = 20;
        CBeta = 30;
        neutroProb = 0;
        
        spike = paramMin;
    }
    
    /**Sets chosen parameter value to correct increment depending on batchNum*/
    void setParameterValue(){
        batchNum = (int) this.job()/numRuns;
        spike+= batchNum*paramInc;
    }
    
    @Override
    public void start() {
        super.start();
        agentsEnvironment = new Continuous2D(5,(XMAX-XMIN),(YMAX-YMIN));
        resetValues();
        setParameterValue();
        
        //Create output CSV files with numbered filenames
        String filename = "model-output"+this.job()+".csv";
        String macroFilename = "macro-output"+this.job()+".csv";
        
        try {
		outFile = new PrintWriter(new FileWriter(filename, true));
                macroOutFile = new PrintWriter(new FileWriter(macroFilename, true));
	} catch (IOException e) {
		e.printStackTrace();
	}
        
        //Create Neutro agents
        for (int i = 0; i < startNeutros; i++) {
            makeNeutro();
        }
        
        //Print column headers for output files
        outFile.println("time, numNeutros, numLive, numApop, numNecrotic, neutroProb(%), CBeta, CAlpha");
        macroOutFile.print("time,");
        
        //Create Macro agents
        for (int i = 0; i < numMacros; i++) {
            makeMacro(i);
        }
        
        schedule.setShuffling(false);
        
        //Create StemCell agent: controls introduction of new Neutro agents and tests for end of simulation
        StemCell stem = new StemCell();
        schedule.scheduleRepeating(stem);
    }
    
    //**Creates and schedules a new Neutro agent with random position and initial direction*/
    public void makeNeutro() {
        double xpos = random.nextDouble()*(XMAX-XMIN);
        double ypos = random.nextDouble()*(YMAX-YMIN);
        double angle = random.nextDouble()*2*Math.PI;
        int life = (int) (300*Math.abs((1+random.nextGaussian())));;
        Neutro neutro = new Neutro(velocity,angle,life);
        setObjectLocation(neutro, new Double2D(xpos,ypos));
        schedule.scheduleRepeating(neutro);
        numNeutros++;
        numLive++;
    }
    
    /**Creates a new Macro agent and writes column headers for it*/
    public void makeMacro(int id) {
        double xpos = random.nextDouble()*(XMAX-XMIN);
        double ypos = random.nextDouble()*(YMAX-YMIN);
        double angle = random.nextDouble()*360;
        Macro wander = new Macro(velocity,angle, id, PAMax, PBMax);
        setObjectLocation(wander, new Double2D(xpos,ypos));
        schedule.scheduleRepeating(wander);
        macroOutFile.print("numApopEaten"+id+",numNecEaten"+id+",RApop"+id+",RNec"+id+",RAlpha"+id+",RBeta"+id+",PAlpha"+id+",PBeta"+id+",");
    }
    
    /**Calculates chemotactic change in direction of Macro agent: attracted towards nearby edible agents
     * and repelled by other Macro agents*/
    public void findNewDirection(Macro macro) {
        double dx;
        double dy;
        //Momentum of current direction
        double vectorX = Math.cos(macro.angle);
        double vectorY = Math.sin(macro.angle);
        macro.necRecord[macro.age]=0;
        macro.apopRecord[macro.age]=0;
        tempAngle = macro.angle;
        Bag nearbyAgents = agentsEnvironment.getObjectsWithinDistance(macro.loc, macro.diam*5);
        if (nearbyAgents != null) {
            for (int i = 0; i < nearbyAgents.numObjs; i++) {
                if (nearbyAgents.objs[i] != macro) {
                    Agent neighbour = (Agent) (nearbyAgents.objs[i]);
                    if (neighbour.isaPhage && neighbour.loc != macro.loc) {
                        dx = macro.loc.x - neighbour.loc.x;
                        dy = macro.loc.y - neighbour.loc.y;
                        //0.1 to prevent divide by zero errors
                        vectorX += dx/(0.1+dx*dx+dy*dy);
                        vectorY += dy/(0.1+dx*dx+dy*dy);
                    }
                    else if (neighbour.necrotic && neighbour.loc != macro.loc) {
                        macro.necRecord[macro.age]++;
                        dx = neighbour.loc.x - macro.loc.x;
                        dy = neighbour.loc.y - macro.loc.y;
                        vectorX += dx/(0.1+dx*dx+dy*dy);
                        vectorY += dy/(0.1+dx*dx+dy*dy);
                    }
                    else if (neighbour.edible && neighbour.loc != macro.loc) {
                        macro.apopRecord[macro.age]++;
                        dx = neighbour.loc.x - macro.loc.x;
                        dy = neighbour.loc.y - macro.loc.y;
                        vectorX += dx/(0.1+dx*dx+dy*dy);
                        vectorY += dy/(0.1+dx*dx+dy*dy);
                    }
                }
            }
            tempAngle = Math.atan2(vectorY, vectorX);
        }
        //Change angle of movement
        macro.angle = tempAngle;
    }
    
    /**Method to handle collisions between two agents: either phagocytosis or bounce off each other*/
    public void collision (final Agent agent) {
        Bag agentBag = agentsEnvironment.getObjectsWithinDistance(agent.loc, agent.diam);
        if ( agentBag != null) {
            for (int i = 0; i < agentBag.numObjs ; i++) {
                if (agentBag.objs[i] != null && agentBag.objs[i] != agent) {
                    Agent bouncee = (Agent) (agentBag.objs[i]);
                    if (agent.isaPhage && bouncee.edible) {
                        bouncee.dead = true;
                        agentsEnvironment.remove(bouncee);
                        //Phagocytosis of necrotic cell
                        if (bouncee.necrotic) {
                            numNeutros--;
                            numNecrotic--;
                            agent.numNecEaten++;
                            bouncee = null;
                            break;
                        }
                        //Phagocytosis of opsonised apoptotic cell
                        else if (bouncee.edible) {
                            numNeutros--;
                            numBound--;
                            agent.numApopEaten++;
                            bouncee = null;
                            break;
                        }
                    }
                    //Cells bounce off each other
                    if (Math.pow(agent.loc.x - bouncee.loc.x,2)
                    +Math.pow(agent.loc.y - bouncee.loc.y,2) < Math.pow(agent.diam + bouncee.diam,2)/2)
                    {
                        tempAngle = agent.angle;
                        agent.angle = bouncee.angle;
                        bouncee.angle = tempAngle;
                        return;
                    }
                }
            }
        }
        return;
    }
    
    /**Method to adjust concentration of signalling molecules (called by Macro agents)*/
    void adjustConcentration(double amountBeta, double amountAlpha) {
        CBeta += amountBeta;
        CAlpha += amountAlpha;
    }
    
    void spikeConcentration() {
        CBeta+=spike;
    }
    
    /**Writes model data to outFile and Macro agent data from macroIntData
     and macroDoubleData to macroOutFile. Called by StemCell agent.*/
    void writeOutputData() {
        int ticks = (int) schedule.time();
        //Model data
        outFile.print(ticks + "," + numNeutros + "," + numLive + "," + (numApop+numBound) + "," 
                + numNecrotic + "," + neutroProb*100 + "," + CBeta + "," + CAlpha + "\n");
        //Macro agent data
        macroOutFile.print("\n"+ticks+",");
        for (int i = 0; i < 20; i++) {
            macroOutFile.print(macroIntData[0][i]+",");
            macroOutFile.print(macroIntData[1][i]+",");
            for (int j = 0; j < 6; j++) {
                macroOutFile.print(macroDoubleData[j][i]+",");
            }

        }
        
    }
    
    /**Method to write data from Macro agents into two arrays at certain timesteps.
     Called by Macro agents.*/
    void writeMacroData(int id, int numApopEaten, int numNecEaten, double RApop, double RNec, double RAlpha, double RBeta, double PAlpha, double PBeta) {
        macroIntData[0][id] = numApopEaten;
        macroIntData[1][id] = numNecEaten;
        macroDoubleData[0][id] = RApop;
        macroDoubleData[1][id] = RNec;
        macroDoubleData[2][id] = RAlpha;
        macroDoubleData[3][id] = RBeta;
        macroDoubleData[4][id] = PAlpha;
        macroDoubleData[5][id] = PBeta;
        return;
    }
    
    /**Ends simulation. Called by StemCell agent after 5000 timesteps*/
    public void finish() {
        super.finish();
	outFile.close();
    }
    
    /**Main method*/
    public static void main(String[] args) {
        doLoop(Engine.class, args);
        System.exit(0);
    }
    
}